##############################################################################
#
# Copyright (c) 2015 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

# pylint:disable=inherit-non-class,no-self-argument,redefined-builtin,c-extension-no-member
# pylint:disable=protected-access

from zope.interface import Interface
from zope.interface import implementer

from persistent import _ring

class IRing(Interface):
    """Conceptually, a doubly-linked list for efficiently keeping track of least-
    and most-recently used :class:`persistent.interfaces.IPersistent` objects.

    This is meant to be used by the :class:`persistent.picklecache.PickleCache`
    and should not be considered a public API. This interface documentation exists
    to assist development of the picklecache and alternate implementations by
    explaining assumptions and performance requirements.
    """

    def __len__(): # pylint:disable=no-method-argument
        """Return the number of persistent objects stored in the ring.

        Should be constant time.
        """

    def __contains__(object): # pylint:disable=unexpected-special-method-signature
        """Answer whether the given persistent object is found in the ring.

        Must not rely on object equality or object hashing, but only
        identity or the `_p_oid`. Should be constant time.
        """

    def add(object):
        """Add the persistent object to the ring as most-recently used.

        When an object is in the ring, the ring holds a strong
        reference to it so it can be deactivated later by the pickle
        cache. Should be constant time.

        The object should not already be in the ring, but this is not necessarily
        enforced.
		"""

    def delete(object):
        """Remove the object from the ring if it is present.

        Returns a true value if it was present and a false value
        otherwise. An ideal implementation should be constant time,
        but linear time is allowed.
        """

    def move_to_head(object):
        """Place the object as the most recently used object in the ring.

        The object should already be in the ring, but this is not
        necessarily enforced, and attempting to move an object that is
        not in the ring has undefined consequences. An ideal
        implementation should be constant time, but linear time is
        allowed.
        """

    def __iter__(): # pylint:disable=no-method-argument
        """Iterate over each persistent object in the ring, in the order of least
        recently used to most recently used.

        Mutating the ring while an iteration is in progress has
        undefined consequences.
        """


ffi = _ring.ffi
_FFI_RING = _ring.lib

_OGA = object.__getattribute__
_OSA = object.__setattr__

_handles = set()

@implementer(IRing)
class _CFFIRing:
    """A ring backed by a C implementation. All operations are constant time.

    It is only available on platforms with ``cffi`` installed.
    """

    __slots__ = ('ring_home', 'ring_to_obj', 'cleanup_func')

    def __init__(self, cleanup_func=None):
        node = self.ring_home = ffi.new("CPersistentRing*")
        node.r_next = node
        node.r_prev = node

        self.cleanup_func = cleanup_func

        # The Persistent objects themselves are responsible for keeping
        # the CFFI nodes alive, but we need to be able to detect whether
        # or not any given object is in our ring, plus know how many there are.
        # In addition, once an object enters the ring, it must be kept alive
        # so that it can be deactivated.
        # Note that because this is a strong reference to the
        # persistent object, its cleanup function --- triggered by the ``ffi.gc``
        # object it owns --- will never be fired while it is in this dict.
        self.ring_to_obj = {}

    def ring_node_for(self, persistent_object, create=True):
        ring_data = _OGA(persistent_object, '_Persistent__ring')
        if ring_data is None:
            if not create:
                return None

            if self.cleanup_func:
                node = ffi.new('CPersistentRingCFFI*')
                node.pobj_id = ffi.cast('uintptr_t', id(persistent_object))
                gc_ptr = ffi.gc(node, self.cleanup_func)
            else:
                node = ffi.new("CPersistentRing*")
                gc_ptr = None
            ring_data = (
                node,
                gc_ptr,
            )
            _OSA(persistent_object, '_Persistent__ring', ring_data)

        return ring_data[0]

    def __len__(self):
        return len(self.ring_to_obj)

    def __contains__(self, pobj):
        node = self.ring_node_for(pobj, False)
        return node and node in self.ring_to_obj

    def add(self, pobj):
        node = self.ring_node_for(pobj)
        _FFI_RING.cffi_ring_add(self.ring_home, node)
        self.ring_to_obj[node] = pobj

    def delete(self, pobj):
        its_node = self.ring_node_for(pobj, False)
        our_obj = self.ring_to_obj.pop(its_node, self)
        if its_node is not None and our_obj is not self and its_node.r_next:
            _FFI_RING.cffi_ring_del(its_node)
            return 1
        return None

    def delete_node(self, node):
        # Minimal sanity checking, assumes we're called from iter.
        self.ring_to_obj.pop(node)
        _FFI_RING.cffi_ring_del(node)

    def move_to_head(self, pobj):
        node = self.ring_node_for(pobj, False)
        _FFI_RING.cffi_ring_move_to_head(self.ring_home, node)

    def iteritems(self):
        head = self.ring_home
        here = head.r_next
        ring_to_obj = self.ring_to_obj
        while here != head:
            # We allow mutation during iteration, which means
            # we must get the next ``here`` value before
            # yielding, just in case the current value is
            # removed.
            current = here
            here = here.r_next
            pobj = ring_to_obj[current]
            yield current, pobj


    def __iter__(self):
        for _, v in self.iteritems():
            yield v

# Export the best available implementation
Ring = _CFFIRing



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Prefer C implementations of Persistent / PickleCache / TimeStamp.

Fall back to pure Python implementations.
"""

import sys

__all__ = [
    'IPersistent',
    'Persistent',
    'GHOST',
    'UPTODATE',
    'CHANGED',
    'STICKY',
    'PickleCache',
    'TimeStamp',
]

# Take care not to shadow the module names
from persistent import interfaces as _interfaces
from persistent import timestamp as _timestamp
from persistent import persistence as _persistence
from persistent import picklecache as _picklecache

IPersistent = _interfaces.IPersistent
Persistent = _persistence.Persistent
GHOST = _interfaces.GHOST
UPTODATE = _interfaces.UPTODATE
CHANGED = _interfaces.CHANGED
STICKY = _interfaces.STICKY
PickleCache = _picklecache.PickleCache

# BWC for TimeStamp.
TimeStamp = _timestamp

sys.modules['persistent.TimeStamp'] = sys.modules['persistent.timestamp']



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Python implementation of persistent base types."""
import persistent
from collections import UserDict as IterableUserDict

class default:

    def __init__(self, func):
        self.func = func

    def __get__(self, inst, class_):
        if inst is None:
            return self
        return self.func(inst)


class PersistentMapping(IterableUserDict, persistent.Persistent):
    """A persistent wrapper for mapping objects.

    This class allows wrapping of mapping objects so that object
    changes are registered.  As a side effect, mapping objects may be
    subclassed.

    A subclass of PersistentMapping or any code that adds new
    attributes should not create an attribute named _container.  This
    is reserved for backwards compatibility reasons.
    """

    # UserDict provides all of the mapping behavior.  The
    # PersistentMapping class is responsible marking the persistent
    # state as changed when a method actually changes the state.  At
    # the mapping API evolves, we may need to add more methods here.

    __super_delitem = IterableUserDict.__delitem__
    __super_setitem = IterableUserDict.__setitem__
    __super_clear = IterableUserDict.clear
    __super_update = IterableUserDict.update
    __super_setdefault = IterableUserDict.setdefault
    __super_pop = IterableUserDict.pop
    __super_popitem = IterableUserDict.popitem


    # Be sure to make a deep copy of our ``data`` (See PersistentList.)
    # See https://github.com/python/cpython/commit/3645d29a1dc2102fdb0f5f0c0129ff2295bcd768
    # This was fixed in CPython 3.7.4, but we can't rely on that because it
    # doesn't handle our old ``_container`` appropriately (it goes directly
    # to ``self.__dict__``, bypassing the descriptor). The code here was initially
    # based on the version found in 3.7.4.
    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        if '_container' in inst.__dict__:
            # BWC for ZODB < 3.3.
            data = inst.__dict__.pop('_container')
        else:
            data = inst.__dict__['data']
        inst.__dict__["data"] = data.copy()
        return inst

    def __delitem__(self, key):
        self.__super_delitem(key)
        self._p_changed = 1

    def __setitem__(self, key, v):
        self.__super_setitem(key, v)
        self._p_changed = 1

    def clear(self):
        """
        Remove all data from this dictionary.

        .. versionchanged:: 4.5.2
           If there was nothing to remove, this object is no
           longer marked as modified.
        """
        # Historically this method always marked ourself as changed,
        # so if there was a _container it was persisted as data. We want
        # to preserve that, even if we won't make any modifications otherwise.
        needs_changed = '_container' in self.__dict__ or bool(self)
        # Python 2 implements this by directly calling self.data.clear(),
        # but Python 3 does so by repeatedly calling self.popitem()
        self.__super_clear()
        if needs_changed:
            self._p_changed = 1

    def update(self, *args, **kwargs):
        """
        D.update([E, ]**F) -> None.

        .. versionchanged:: 4.5.2
           Now accepts arbitrary keyword arguments. In the special case
           of a keyword argument named ``b`` that is a dictionary,
           the behaviour will change.
        """
        self.__super_update(*args, **kwargs)
        self._p_changed = 1

    def setdefault(self, key, default=None):
        # We could inline all of UserDict's implementation into the
        # method here, but I'd rather not depend at all on the
        # implementation in UserDict (simple as it is).
        if key not in self.data:
            self._p_changed = 1
        return self.__super_setdefault(key, default=default)

    def pop(self, key, *args, **kwargs):
        self._p_changed = 1
        return self.__super_pop(key, *args, **kwargs)

    def popitem(self):
        """
        Remove an item.

        .. versionchanged:: 4.5.2
           No longer marks this object as modified if it was empty
           and an exception raised.
        """
        result = self.__super_popitem()
        self._p_changed = 1
        return result

    # Old implementations (prior to 2001; see
    # https://github.com/zopefoundation/ZODB/commit/c64281cf2830b569eed4f211630a8a61d22a0f0b#diff-b0f568e20f51129c10a096abad27c64a)
    # used ``_container`` rather than ``data``. Use a descriptor to provide
    # ``data`` when we have ``_container`` instead

    @default
    def data(self): # pylint:disable=method-hidden
        # We don't want to cause a write on read, so we're careful not to
        # do anything that would cause us to become marked as changed, however,
        # if we're modified, then the saved record will have data, not
        # _container.
        data = self.__dict__.pop('_container')
        self.__dict__['data'] = data

        return data



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Python implementation of persistent list."""
import sys
import persistent
from collections import UserList

# The slice object you get when you write list[:]
_SLICE_ALL = slice(None, None, None)


class PersistentList(UserList, persistent.Persistent):
    """A persistent wrapper for list objects.

    Mutating instances of this class will cause them to be marked
    as changed and automatically persisted.

    .. versionchanged:: 4.5.2
       Using the `clear` method, or deleting a slice (e.g., ``del inst[:]`` or ``del inst[x:x]``)
       now only results in marking the instance as changed if it actually removed
       items.
    """
    __super_getitem = UserList.__getitem__
    __super_setitem = UserList.__setitem__
    __super_delitem = UserList.__delitem__
    __super_iadd = UserList.__iadd__
    __super_imul = UserList.__imul__
    __super_append = UserList.append
    __super_insert = UserList.insert
    __super_pop = UserList.pop
    __super_remove = UserList.remove
    __super_reverse = UserList.reverse
    __super_sort = UserList.sort
    __super_extend = UserList.extend
    __super_clear = (
        UserList.clear
        if hasattr(UserList, 'clear')
        else lambda inst: inst.__delitem__(_SLICE_ALL)
    )

    if sys.version_info[:3] < (3, 7, 4):  # pragma: no cover
        # Prior to 3.7.4, Python 3 failed to properly
        # return an instance of the same class.
        # See https://bugs.python.org/issue27639
        # and https://github.com/zopefoundation/persistent/issues/112.
        # We only define the special method on the necessary versions to avoid
        # any speed penalty.
        def __getitem__(self, item):
            result = self.__super_getitem(item)
            if isinstance(item, slice):
                result = self.__class__(result)
            return result

    if sys.version_info[:3] < (3, 7, 4):  # pragma: no cover
        # Likewise for __copy__.
        # See https://github.com/python/cpython/commit/3645d29a1dc2102fdb0f5f0c0129ff2295bcd768
        def __copy__(self):
            inst = self.__class__.__new__(self.__class__)
            inst.__dict__.update(self.__dict__)
            # Create a copy and avoid triggering descriptors
            inst.__dict__["data"] = self.__dict__["data"][:]
            return inst

    def __setitem__(self, i, item):
        self.__super_setitem(i, item)
        self._p_changed = 1

    def __delitem__(self, i):
        # If they write del list[:] but we're empty,
        # no need to mark us changed. Likewise with
        # a slice that's empty, like list[1:1].
        len_before = len(self.data)
        self.__super_delitem(i)
        if len(self.data) != len_before:
            self._p_changed = 1

    def __iadd__(self, other):
        L = self.__super_iadd(other)
        self._p_changed = 1
        return L

    def __imul__(self, n):
        L = self.__super_imul(n)
        self._p_changed = 1
        return L

    def append(self, item):
        self.__super_append(item)
        self._p_changed = 1

    def clear(self):
        """
        Remove all items from the list.

        .. versionchanged:: 4.5.2
           Now marks the list as changed.
        """
        needs_changed = bool(self)
        self.__super_clear()
        if needs_changed:
            self._p_changed = 1

    def insert(self, i, item):
        self.__super_insert(i, item)
        self._p_changed = 1

    def pop(self, i=-1):
        rtn = self.__super_pop(i)
        self._p_changed = 1
        return rtn

    def remove(self, item):
        self.__super_remove(item)
        self._p_changed = 1

    def reverse(self):
        self.__super_reverse()
        self._p_changed = 1

    def sort(self, *args, **kwargs):
        self.__super_sort(*args, **kwargs)
        self._p_changed = 1

    def extend(self, other):
        self.__super_extend(other)
        self._p_changed = 1
